/*
 * bn256-20080525/bnpairings.c 
 * Peter Schwabe
 * Public domain
 */
#ifdef PYTHON
#include <Python.h>
#include <errno.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#ifdef TEST
#include <assert.h>
#endif

#include "init.h"
#include "curvepoint_fp.h"
#include "twistpoint_fp2.h"
#include "points.h"
#include "fp12e.h"
#include "comfp12e.h"
#include "randompoints.h"

#include "ate_optate.h"
#include "eta_tate.h"
#include "cometa_comtate.h"

#ifdef TEST
int main(int args,char* argv[])
{
  assert(args== 3);
  
  init_globals();

  curvepoint_fp_t point1;
  twistpoint_fp2_t point2;
    
  curvepoint_fp_set(point1, curve_gen);
  twistpoint_fp2_set(point2, twist_gen);
  comfp12e_t rop;
  comtate_param(rop, point1, point2, atoi(argv[1]), atoi(argv[2]));
  comfp12e_makeaffine(rop, rop);

  static char* output[5];
  comfp12e_output(output,rop);
  
  return 0;
}
#endif



#ifdef PYTHON

char** c_tate(int ml, int fe){
  init_globals();

  int i;
  
  curvepoint_fp_t point1;
  twistpoint_fp2_t point2;
    
  curvepoint_fp_set(point1, curve_gen);
  twistpoint_fp2_set(point2, twist_gen);

  comfp12e_t rop;
  comtate_param(rop, point1, point2, ml, fe);

  static char *output[5];
  for (i = 0; i < 5; i++){
    output[i] = malloc(78);
  }
  comfp12e_output(output,rop);

  return output;
}

char** c_puiss(const char* base_str0, const char *base_str1, const char *base_str2, const char *base_str3, const char *base_str4, const char* exp_str){
  int i;
  init_globals();
  
  comfp12e_t base,rop;
  mpz_t exp;


  mpz_init(exp);


  const char *base_str[5] = {base_str0,base_str1,base_str2,base_str3,base_str4};
  comfp12e_set_str(base,base_str);
  mpz_set_str(exp,exp_str,10);
  
  comfp12e_pow(rop,base,exp);

  static char * output[5];
  for (i = 0; i < 5; i++){
    output[i] = malloc(78);
  }
  comfp12e_output(output,rop);

  return output;
}

char** c_invert(const char* base_str0, const char *base_str1, const char *base_str2, const char *base_str3, const char *base_str4){
  init_globals();
  int i;
  comfp12e_t base,rop;

  const char *base_str[5] = {base_str0,base_str1,base_str2,base_str3,base_str4};
  comfp12e_set_str(base,base_str);
  
  comfp12e_invert(rop,base);

  static char * output[5];
  for (i = 0; i < 5; i++){
    output[i] = malloc(78);
  }
  comfp12e_output(output,rop);

  return output;
}

char** c_mul(const char* base_str00, const char *base_str01, const char *base_str02, const char *base_str03, const char *base_str04,
	     const char* base_str10, const char *base_str11, const char *base_str12, const char *base_str13, const char *base_str14){
  int i;
  init_globals();
  
  comfp12e_t base0,rop,base1;

  const char *base0_str[5] = {base_str00,base_str01,base_str02,base_str03,base_str04};
  comfp12e_set_str(base0,base0_str);
  
  const char *base1_str[5] = {base_str10,base_str11,base_str12,base_str13,base_str14};
  comfp12e_set_str(base1,base1_str);
  
  comfp12e_mul(rop,base0,base1);

  static char * output[5];
  for (i = 0; i < 5; i++){
    output[i] = malloc(78);
  }
  comfp12e_output(output,rop);

  return output;
}

char** c_square(const char* base_str0, const char *base_str1, const char *base_str2, const char *base_str3, const char *base_str4){
  init_globals();
  int i;
  comfp12e_t base,rop;

  const char *base_str[5] = {base_str0,base_str1,base_str2,base_str3,base_str4};
  comfp12e_set_str(base,base_str);
  
  comfp12e_square(rop,base);

  static char * output[5];
  for (i = 0; i < 5; i++){
    output[i] = malloc(78);
  }
  comfp12e_output(output,rop);

  return output;
}


char** c_normal_tate(int ml, int fe){
  init_globals();

  int i;
  
  curvepoint_fp_t point1;
  twistpoint_fp2_t point2;
    
  curvepoint_fp_set(point1, curve_gen);
  twistpoint_fp2_set(point2, twist_gen);

  fp12e_t rop;
  tate_param(rop, point1, point2, ml, fe);
  static char *output[12];
  for (i = 0; i < 12; i++){
    output[i] = malloc(78);
  }
  fp12e_output(output,rop);

  return output;
}

char** c_compress(const char* str0, const char* str1, const char* str2,
		  const char* str3, const char* str4, const char* str5,
		  const char* str6, const char* str7, const char* str8,
		  const char* str9, const char* str10, const char* str11)
{
  //TODO: verify the entred value is a (p^6-1)(p^2+1) power of an element
  const char *str[12] = {str0, str1, str2, str3, str4, str5, str6, str7, str8, str9, str10, str11};
  int i;
  fp12e_t myfp12e;
  comfp12e_t mycomfp12e;
  static char *output[5];
  for (i = 0; i < 5; i++){
    output[i] = malloc(78);
  }
  fp12e_set_str(myfp12e, str);
  fp12e_to_comfp12e(mycomfp12e,myfp12e);
  comfp12e_output(output, mycomfp12e);
  return output;  
}

char** c_normal_puiss(const char* str0, const char* str1, const char* str2,
		      const char* str3, const char* str4, const char* str5,
		      const char* str6, const char* str7, const char* str8,
		      const char* str9, const char* str10, const char* str11,
		      const char* exp_str)
{
  init_globals();
  int i;
  fp12e_t base,rop;
  mpz_t exp;

  mpz_init(exp);
  

  const char *base_str[12] = {str0,str1,str2,str3,str4,str5,str6,str7,str8,str9,str10,str11};
  fp12e_set_str(base,base_str);
  mpz_set_str(exp,exp_str,10);
  
  fp12e_pow(rop,base,exp);

  static char * output[12];
  for (i = 0; i < 12; i++){
    output[i] = malloc(78);
  }
  
  fp12e_output(output,rop);

  return output;
}


char** c_normal_invert(const char* str0, const char* str1, const char* str2,
		       const char* str3, const char* str4, const char* str5,
		       const char* str6, const char* str7, const char* str8,
		       const char* str9, const char* str10, const char* str11)
{
  int i;
  init_globals();
  fp12e_t rop;

  const char *rop_str[12] = {str0,str1,str2,str3,str4,str5,str6,str7,str8,str9,str10,str11};
  fp12e_set_str(rop,rop_str);
  
  fp12e_invert(rop,rop);

  static char * output[12];
  for (i = 0; i < 12; i++){
    output[i] = malloc(78);
  }
  
  fp12e_output(output,rop);

  return output;
}

static PyObject *
bn256_c_tate(PyObject *self, PyObject *args) {
  assert(self == NULL);
  int ml=0;
  int fe=0;
  int i;
  char ** result;
  
  if (!PyArg_ParseTuple(args, "ii", &ml, &fe))
    return NULL;
  
  result = c_tate(ml,fe);

  PyObject *lst = PyList_New(5);
  if (!lst)
    return NULL;
  for (i = 0; i < 5; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}


static PyObject *
bn256_c_puiss(PyObject *self, PyObject *args) {
  assert(self == NULL);

  int i;
  char *str0, *str1, *str2, *str3, *str4, *exp; 
  char **result;
  
  if (!PyArg_ParseTuple(args, "ssssss", &str0, &str1, &str2, &str3, &str4, &exp))
    return NULL;

  result = c_puiss(str0,str1,str2,str3,str4,exp);

  PyObject *lst = PyList_New(5);
  if (!lst)
    return NULL;
  for (i = 0; i < 5; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}

static PyObject *
bn256_c_invert(PyObject *self, PyObject *args) {
  assert(self == NULL);

  int i;
  char *str0, *str1, *str2, *str3, *str4; 
  char **result;
  
  if (!PyArg_ParseTuple(args, "sssss", &str0, &str1, &str2, &str3, &str4))
    return NULL;

  result = c_invert(str0,str1,str2,str3,str4);

  PyObject *lst = PyList_New(5);
  if (!lst)
    return NULL;
  for (i = 0; i < 5; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}


static PyObject *
bn256_c_mul(PyObject *self, PyObject *args) {
  assert(self == NULL);

  int i;
  char *str00, *str01, *str02, *str03, *str04, *str10, *str11, *str12, *str13, *str14; 
  char **result;
  
  if (!PyArg_ParseTuple(args, "ssssssssss", &str00, &str01, &str02, &str03, &str04, &str10, &str11, &str12, &str13, &str14))
    return NULL;

  result = c_mul(str00,str01,str02,str03,str04,str10,str11,str12,str13,str14);

  PyObject *lst = PyList_New(5);
  if (!lst)
    return NULL;
  for (i = 0; i < 5; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}


static PyObject *
bn256_c_square(PyObject *self, PyObject *args) {
  assert(self == NULL);

  int i;
  char *str0, *str1, *str2, *str3, *str4; 
  char **result;
  
  if (!PyArg_ParseTuple(args, "sssss", &str0, &str1, &str2, &str3, &str4))
    return NULL;

  result = c_square(str0,str1,str2,str3,str4);

  PyObject *lst = PyList_New(5);
  if (!lst)
    return NULL;
  for (i = 0; i < 5; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst; 
}

static PyObject *
bn256_c_normal_tate(PyObject *self, PyObject *args) {
  assert(self == NULL);
  int ml=0;
  int fe=0;
  int i;
  char ** result;
  
  if (!PyArg_ParseTuple(args, "ii", &ml, &fe))
    return NULL;
  
  result = c_normal_tate(ml,fe);

  PyObject *lst = PyList_New(12);
  if (!lst)
    return NULL;
  for (i = 0; i < 12; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}

static PyObject *
bn256_c_compress(PyObject *self, PyObject *args) {
  assert(self == NULL);

  int i;
  char *str0, *str1, *str2, *str3, *str4, *str5, *str6, *str7, *str8, *str9,*str10, *str11; 
  char **result;
  
  if (!PyArg_ParseTuple(args, "ssssssssssss", &str0, &str1, &str2, &str3, &str4, &str5, &str6, &str7, &str8, &str9, &str10, &str11))
    return NULL;

  result = c_compress(str0,str1,str2,str3,str4,str5,str6,str7,str8,str9,str10,str11);

  PyObject *lst = PyList_New(5);
  if (!lst)
    return NULL;
  for (i = 0; i < 5; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}

static PyObject *
bn256_c_normal_puiss(PyObject *self, PyObject *args) {
  assert(self == NULL);

  int i;
  char *str0, *str1, *str2, *str3, *str4, *str5, *str6, *str7, *str8, *str9, *str10, *str11, *exp; 
  char **result;
  
  if (!PyArg_ParseTuple(args, "sssssssssssss", &str0, &str1, &str2, &str3, &str4, &str5, &str6, &str7, &str8, &str9, &str10, &str11, &exp))
    return NULL;

  result = c_normal_puiss(str0,str1,str2,str3,str4,str5,str6,str7,str8,str9,str10,str11,exp);

  PyObject *lst = PyList_New(12);
  if (!lst)
    return NULL;
  for (i = 0; i < 12; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}

static PyObject *
bn256_c_normal_invert(PyObject *self, PyObject *args) {
  assert(self == NULL);

  int i;
  char *str0, *str1, *str2, *str3, *str4, *str5, *str6, *str7, *str8, *str9, *str10, *str11; 
  char **result;
  
  if (!PyArg_ParseTuple(args, "ssssssssssss", &str0, &str1, &str2, &str3, &str4, &str5, &str6, &str7, &str8, &str9, &str10, &str11))
    return NULL;

  result = c_normal_invert(str0,str1,str2,str3,str4,str5,str6,str7,str8,str9,str10,str11);

  PyObject *lst = PyList_New(12);
  if (!lst)
    return NULL;
  for (i = 0; i < 12; i++) {
    PyObject *num = PyString_FromString(result[i]);
    if (!num) {
        Py_DECREF(lst);
        return NULL;
    }
    PyList_SET_ITEM(lst, i, num);   // reference to num stolen
  }
  return lst;
}


static PyMethodDef
bn256Methods[] = {
  {"tate", bn256_c_tate, METH_VARARGS,"returns the result of the calculate of tate-pairing"},
  {"puiss", bn256_c_puiss, METH_VARARGS,"returns the result of the base^exp"},
  {"invert",bn256_c_invert,METH_VARARGS,"retruns the invert of the argument"},
  {"mul",bn256_c_mul,METH_VARARGS,"retruns the multiplication of the two arguments"},
  {"square",bn256_c_square,METH_VARARGS,"retruns the square of the arguments"},
  {"normal_tate", bn256_c_normal_tate, METH_VARARGS,"returns the result of the calculate of tate-pairing wihout compression"},
  {"compress",bn256_c_compress, METH_VARARGS,"returns the compressed form of the pairing's intermediate value"},
  {"normal_puiss",bn256_c_normal_puiss, METH_VARARGS,"returns the power of the given argument"},
  {"normal_invert",bn256_c_normal_invert, METH_VARARGS,"returns the invert of the given argument"},

  {NULL, NULL, 0,NULL},
};

void initbn256() {
  PyImport_AddModule("bn256");
  Py_InitModule("bn256", bn256Methods);
}

int main(int argc, char **argv)
{
        /* Pass argv[0] to the Python interpreter */
        Py_SetProgramName(argv[0]);

        /* Initialize the Python interpreter.  Required. */
        Py_Initialize();

        /* Add a static module */
        initbn256();

        /* Define sys.argv.  It is up to the application if you
           want this; you can also let it undefined (since the Python
           code is generally not a main program it has no business
           touching sys.argv...) */
        PySys_SetArgv(argc, argv);

        /* Exit, cleaning up the interpreter */
        Py_Exit(0);
        /*NOTREACHED*/
	return 0;
}



#endif
